
#include <sys/p2es_clock.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))


enum {
	_clkfreq = 20_000_000,
};



static uint32_t  div33( uint32_t dividend, uint32_t divisor )
{
	uint32_t  x;

	__asm {
		mov	x, divisor
		shr	x, #1
		add	dividend, x
		qdiv	dividend, divisor
		getqx	x
	}

	return  x;
}


static uint32_t  muldiv65( uint32_t mult1, uint32_t mult2, uint32_t divisor )
{
	uint32_t  x;

	__asm {
		qmul	mult1, mult2
		mov	x, divisor
		shr	x, #1
		getqx	mult1
		getqy	mult2
		add	mult1, x    wc
		addx	mult2, #0
		setq	mult2
		qdiv	mult1, divisor
		getqx	x
	}

	return  x;
}


static uint32_t  mul2div65( uint16_t mult1, uint16_t mult2, uint32_t mult32, uint32_t divisor )
{
	uint32_t  x;

	__asm {
		mul	mult1, mult2
		mov	x, mult1
	}

	return  muldiv65( x, mult32, divisor );
}


enum {
	_MINFPFD = 250_000,
};

static uint32_t  pll_calc( uint32_t targetfreq )  // returns a suitable _clkset() mode value, or 0 for an out-of-bounds result
{
	uint32_t  mult, divd, divp, post, Fpfd, Fvco, mode = 0;
	int32_t  error, besterror = _MINFPFD;

	for( post = 0; post <= 15; post++ )
	{
		divp = post ? post * 2 : 1;
		for( divd = 64; divd >= 1; divd-- )
		{
			mult = mul2div65( divp, divd, targetfreq, _XTALFREQ );
			Fvco = muldiv65( _XTALFREQ, mult, divd );
			Fpfd = div33( _XTALFREQ, divd );

			if( (Fpfd >= _MINFPFD)
				&& (mult <= 1024)
				&& (Fvco >= 100_000_000)
				&& ((Fvco <= 201_000_000) || (Fvco <= targetfreq + _MINFPFD)) )
			{
				error = abs( (int32_t)(div33( Fvco, divp ) - targetfreq) );
				if( error <= besterror )
				{
					mode = 1<<24 | (divd-1)<<18 | (mult-1)<<8 | ((post-1)&0x0f)<<4 | 0x0b;  // %CC==%10 %SS==11
					besterror = error;
				}
			}
		}
	}

	return  mode;
}


enum {
	ADCPIN = 8,
	ADCPINS = ADCPIN | (7<<6),
};

__pasm {
startadc
		fltl	#ADCPINS
		wrpin	##P_ADC | P_ADC_100X, #ADCPINS		' turn on ADC modulator and digital filter smartpin mode
		wxpin	#%10_0000, #ADCPINS			' Sinc3 filter sub-mode
		wypin	#500, #ADCPINS				' set decimation period in sysclocks
		waitx	##_clkfreq/100				' 10 ms pause for ADC modulator warm up

		dirh	#ADCPINS				' enable digital filters
'		mov	pa, #ADCPIN
'		rep	@.rend, #(ADCPINS>>6)+1
'		dirh	pa					' enable digital filters, one pin at a time to stagger them
'		add	pa, #1
'		waitx	#4
.rend

		wrfast	##5000, ptra			' 10_000 samples x 8 channels, PTRA is hubRAM address of "mailbox"
		setse1	#%001<<6 | ADCPIN
		rdlong	launcher, ptra			' fetch the stored cog ID of launcher cog
		pollfbw					' clear any prior event
		pollse1					' clear any prior event

		rep	@.done, #0			' repeat forever
		waitse1					' wait for next decimation event
		jfbw	#.done				' break the REP upon FIFO wrapping event

		rdpin	pa, #ADCPIN
		sub	pa, diff1
		add	diff1, pa
		sub	pa, diff2
		add	diff2, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+1
		sub	pa, diff1+1
		add	diff1+1, pa
		sub	pa, diff2+1
		add	diff2+1, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+2
		sub	pa, diff1+2
		add	diff1+2, pa
		sub	pa, diff2+2
		add	diff2+2, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+3
		sub	pa, diff1+3
		add	diff1+3, pa
		sub	pa, diff2+3
		add	diff2+3, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+4
		sub	pa, diff1+4
		add	diff1+4, pa
		sub	pa, diff2+4
		add	diff2+4, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+5
		sub	pa, diff1+5
		add	diff1+5, pa
		sub	pa, diff2+5
		add	diff2+5, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+6
		sub	pa, diff1+6
		add	diff1+6, pa
		sub	pa, diff2+6
		add	diff2+6, pa
		zerox	pa, #26
		wflong	pa

		rdpin	pa, #ADCPIN+7
		sub	pa, diff1+7
		add	diff1+7, pa
		sub	pa, diff2+7
		add	diff2+7, pa
		zerox	pa, #26
		wflong	pa

.done
		dirl	#ADCPINS
		wrpin	#0, #ADCPINS
		decod	launcher
		cogatn	launcher
		cogid	pa
		cogstop	pa
		jmp	#$


launcher	long	0
diff1		res	8
diff2		res	8
}


int  main( void )
{
	int32_t volatile * mailbox;
	unsigned  idx, id;


	_waitms( 200 );						// pause for terminal ready

	mailbox = __builtin_alloca( 5000 * 64 + 8 );		// 10_000 samples x 8 channels
	if( mailbox == NULL )  return( 1 );
//	memset( mailbox, 0, 5000 * 64 + 8 );

	mailbox[0] = _cogid();
	id = _coginit( COGEXEC_NEW, &startadc, mailbox );
	if( id == -1 )  return( 2 );

	_waitatn();

	printf( " P%d P%d P%d P%d P%d P%d P%d P%d\n", ADCPIN, ADCPIN+1, ADCPIN+2, ADCPIN+3, ADCPIN+4, ADCPIN+5, ADCPIN+6, ADCPIN+7 );
	idx = 80;						// start 10 samples in
	do {
		printf( " %d %d %d %d %d %d %d %d\n", mailbox[idx], mailbox[idx+1], mailbox[idx+2], mailbox[idx+3], mailbox[idx+4], mailbox[idx+5], mailbox[idx+6], mailbox[idx+7] );
		idx += 8;
	} while( idx < 80_000 );
	puts( " STOP" );

	return( 0 );
}
